@@ -1,26 +1,23 @@
+# SPDX-FileCopyrightText: 2013 Ole Martin Bjorndalen <ombdalen@gmail.com>
+#
+# SPDX-License-Identifier: MIT
+
 """
 Useful tools for working with ports
-
-Module content:
-
-    multi_receive -- receive messages from multiple ports
-    multi_iter_pending -- iterate through messages from multiple ports
-    IOPort -- combined input / output port. Wraps around to normal ports
-    MessageBuffer -- pseudo-port that stores messages in a deque
 """
-
-from __future__ import unicode_literals
-import time
 import random
-from collections import deque
-from .parser import Parser
+import threading
+import time
+
 from .messages import Message
+from .parser import Parser
 
 # How many seconds to sleep before polling again.
 _DEFAULT_SLEEP_TIME = 0.001
 _sleep_time = _DEFAULT_SLEEP_TIME
 
-# Todo: document this more.
+
+# TODO: document this more.
 def sleep():
     """Sleep for N seconds.
 
@@ -28,22 +25,67 @@ def sleep():
     be set with set_sleep_time()."""
     time.sleep(_sleep_time)
 
+
 def set_sleep_time(seconds=_DEFAULT_SLEEP_TIME):
     """Set the number of seconds sleep() will sleep."""
     global _sleep_time
     _sleep_time = seconds
 
+
 def get_sleep_time():
     """Get number of seconds sleep() will sleep."""
     return _sleep_time
 
-class BasePort(object):
+
+def reset_messages():
+    """Yield "All Notes Off" and "Reset All Controllers" for all channels"""
+    ALL_NOTES_OFF = 123
+    RESET_ALL_CONTROLLERS = 121
+    for channel in range(16):
+        for control in [ALL_NOTES_OFF, RESET_ALL_CONTROLLERS]:
+            yield Message('control_change', channel=channel, control=control)
+
+
+def panic_messages():
+    """Yield "All Sounds Off" for all channels.
+
+    This will mute all sounding notes regardless of
+    envelopes. Useful when notes are hanging and nothing else
+    helps.
+    """
+    ALL_SOUNDS_OFF = 120
+    for channel in range(16):
+        yield Message('control_change',
+                      channel=channel, control=ALL_SOUNDS_OFF)
+
+
+class DummyLock:
+    def __enter__(self):
+        return self
+
+    def __exit__(self, *_):
+        return False
+
+
+class BasePort:
     """
     Abstract base class for Input and Output ports.
     """
+    is_input = False
+    is_output = False
+    _locking = True
 
     def __init__(self, name=None, **kwargs):
+        if hasattr(self, 'closed'):
+            # __init__() called twice (from BaseInput and BaseOutput).
+            # This stops _open() from being called twice.
+            return
+
         self.name = name
+        if self._locking:
+            self._lock = threading.RLock()
+        else:
+            self._lock = DummyLock()
         self.closed = True
         self._open(**kwargs)
         self.closed = False
@@ -54,12 +96,6 @@ class BasePort(object):
     def _close(self):
         pass
 
-    def _send(self, message):
-        pass
-
-    def _receive(self, block=True):
-        pass
-
     def close(self):
         """Close the port.
 
@@ -67,15 +103,16 @@ class BasePort(object):
         is automatically closed when the object goes out of scope or
         is garbage collected.
         """
-        if not self.closed:
-            if hasattr(self, 'autoreset') and self.autoreset:
-                try:
-                    self.reset()
-                except IOError:
-                    pass
+        with self._lock:
+            if not self.closed:
+                if hasattr(self, 'autoreset') and self.autoreset:
+                    try:
+                        self.reset()
+                    except OSError:
+                        pass
 
-            self._close()
-            self.closed = True
+                self._close()
+                self.closed = True
 
     def __del__(self):
         self.close()
@@ -93,13 +130,12 @@ class BasePort(object):
         else:
             state = 'open'
 
-        capabilities = (hasattr(self, 'receive'), hasattr(self, 'send'))
-        port_type = {
-            (True, False): 'input',
-            (False, True): 'output',
-            (True, True): 'I/O',
-            (False, False): 'mute',
-            }[capabilities]
+        capabilities = self.is_input, self.is_output
+        port_type = {(True, False): 'input',
+                     (False, True): 'output',
+                     (True, True): 'I/O port',
+                     (False, False): 'mute port',
+                     }[capabilities]
 
         name = self.name or ''
 
@@ -113,12 +149,12 @@ class BasePort(object):
 
 
 class BaseInput(BasePort):
-    """
-    Base class for input port.
+    """Base class for input port.
 
-    Override _pending() to create a new input port type.
+    Subclass and override _receive() to create a new input port type.
     (See portmidi.py for an example of how to do this.)
     """
+    is_input = True
 
     def __init__(self, name='', **kwargs):
         """Create an input port.
@@ -126,80 +162,86 @@ class BaseInput(BasePort):
         name is the port name, as returned by input_names(). If
         name is not passed, the default input is used instead.
         """
-        self.callback = kwargs.get('callback')
-        self._parser = Parser()
-        self._messages = self._parser._parsed_messages  # Shortcut.
         BasePort.__init__(self, name, **kwargs)
+        self._parser = Parser()
+        self._messages = self._parser.messages  # Shortcut.
 
-    def pending(self):
-        """Return how many messages are ready to be received.
+    def _check_callback(self):
+        if hasattr(self, 'callback') and self.callback is not None:
+            raise ValueError('a callback is set for this port')
 
-        This will read data from the device and put it in the
-        parser. I will then return the number of messages available to
-        be received.
-
-        If this is called on a closed port it will work as normal
-        except it won't try to read from the device.
-        """
-        if not self.closed:
-            self._receive(block=False)
-
-        return len(self._messages)
+    def _receive(self, block=True):
+        pass
 
     def iter_pending(self):
         """Iterate through pending messages."""
-        self._receive(block=False)
-        while self._messages:
-            yield self._messages.popleft()
+        while True:
+            msg = self.poll()
+            if msg is None:
+                return
+            else:
+                yield msg
 
     def receive(self, block=True):
         """Return the next message.
 
-        This will block until a message arrives. For non-blocking
-        behavior, you can use pending() to see how many messages can
-        safely be received without blocking.
+        This will block until a message arrives.
 
-        If the port is closed and there are no pending messages, IOError
+        If you pass block=False it will not block and instead return
+        None if there is no available message.
+
+        If the port is closed and there are no pending messages IOError
         will be raised. If the port closes while waiting inside receive(),
-        IOError will be raised. Todo: this seems a bit inconsistent. Should
+        IOError will be raised. TODO: this seems a bit inconsistent. Should
         different errors be raised? What's most useful here?
-
-        If block=False is passed, None will be returned if there are no
-        pending messages or if the port is closed.
         """
-        if hasattr(self, 'callback') and self.callback:
-            raise IOError('a callback is currently set for this port')
+        if not self.is_input:
+            raise ValueError('Not an input port')
+
+        self._check_callback()
 
         # If there is a message pending, return it right away.
-        if self._messages:
-            return self._messages.popleft()
+        with self._lock:
+            if self._messages:
+                return self._messages.popleft()
 
         if self.closed:
             if block:
-                raise IOError('receive() called on closed port')
+                raise ValueError('receive() called on closed port')
             else:
                 return None
 
-        while 1:
-            self._receive(block=block)
-            if self._messages:
-                return self._messages.popleft()
-            elif not block:
-                return None
-            elif self.closed:
-                raise IOError('port closed during receive()')
+        while True:
+            with self._lock:
+                msg = self._receive(block=block)
+                if msg:
+                    return msg
+
+                if self._messages:
+                    return self._messages.popleft()
+                elif not block:
+                    return None
+                elif self.closed:
+                    raise OSError('port closed during receive()')
 
             sleep()
 
+    def poll(self):
+        """Receive the next pending message or None
+
+        This is the same as calling `receive(block=False)`."""
+        return self.receive(block=False)
+
     def __iter__(self):
         """Iterate through messages until the port closes."""
         # This could have simply called receive() in a loop, but that
         # could result in a "port closed during receive()" error which
         # is hard to catch here.
-        while 1:
+        self._check_callback()
+        while True:
             try:
                 yield self.receive()
-            except IOError:
+            except OSError:
                 if self.closed:
                     # The port closed before or inside receive().
                     # (This makes the assumption that this is the reason,
@@ -208,6 +250,7 @@ class BaseInput(BasePort):
                 else:
                     raise
 
+
 class BaseOutput(BasePort):
     """
     Base class for output port.
@@ -215,42 +258,43 @@ class BaseOutput(BasePort):
     Subclass and override _send() to create a new port type.  (See
     portmidi.py for how to do this.)
     """
+    is_output = True
 
     def __init__(self, name='', autoreset=False, **kwargs):
         """Create an output port
-        
+
         name is the port name, as returned by output_names(). If
         name is not passed, the default output is used instead.
         """
         BasePort.__init__(self, name, **kwargs)
         self.autoreset = autoreset
 
-    def send(self, message):
+    def _send(self, msg):
+        pass
+
+    def send(self, msg):
         """Send a message on the port.
 
         A copy of the message will be sent, so you can safely modify
         the original message without any unexpected consequences.
         """
-        if not isinstance(message, Message):
+        if not self.is_output:
+            raise ValueError('Not an output port')
+        elif not isinstance(msg, Message):
             raise TypeError('argument to send() must be a Message')
         elif self.closed:
             raise ValueError('send() called on closed port')
 
-        self._send(message.copy())
+        with self._lock:
+            self._send(msg.copy())
 
     def reset(self):
-        """Send "All Notes Off" and "Reset All Controllers" on all channels
-        """
+        """Send "All Notes Off" and "Reset All Controllers" on all channels"""
         if self.closed:
             return
 
-        ALL_NOTES_OFF = 123
-        RESET_ALL_CONTROLLERS = 121
-        for channel in range(16):
-            for control in [ALL_NOTES_OFF, RESET_ALL_CONTROLLERS]:
-                self.send(Message('control_change',
-                                  channel=channel,
-                                  control=control))
+        for msg in reset_messages():
+            self.send(msg)
 
     def panic(self):
         """Send "All Sounds Off" on all channels.
@@ -262,14 +306,19 @@ class BaseOutput(BasePort):
         if self.closed:
             return
 
-        ALL_SOUNDS_OFF = 120
-        for channel in range(16):
-            self.send(Message('control_change',
-                              channel=channel,
-                              control=ALL_SOUNDS_OFF))
+        for msg in panic_messages():
+            self.send(msg)
+
 
 class BaseIOPort(BaseInput, BaseOutput):
-    pass
+    def __init__(self, name='', **kwargs):
+        """Create an IO port.
+
+        name is the port name, as returned by ioport_names().
+        """
+        BaseInput.__init__(self, name, **kwargs)
+        BaseOutput.__init__(self, name, **kwargs)
+
 
 class IOPort(BaseIOPort):
     """Input / output port.
@@ -279,24 +328,27 @@ class IOPort(BaseIOPort):
     is forwarded to the appropriate port.
     """
 
+    _locking = False
+
     def __init__(self, input, output):
         self.input = input
         self.output = output
 
         # We use str() here in case name is None.
-        self.name = '{} + {}'.format(str(input.name), str(output.name))
+        self.name = f'{str(input.name)} + {str(output.name)}'
         self._messages = self.input._messages
         self.closed = False
+        self._lock = DummyLock()
 
     def _close(self):
         self.input.close()
         self.output.close()
 
     def _send(self, message):
-        self.output._send(message)
+        self.output.send(message)
 
     def _receive(self, block=True):
-        return self.input._receive()
+        return self.input.receive(block=block)
 
 
 class EchoPort(BaseIOPort):
@@ -306,67 +358,65 @@ class EchoPort(BaseIOPort):
     __iter__ = BaseIOPort.iter_pending
 
 
-# Todo: i don't know how to implement yield_ports here, so for now I haven't.
 class MultiPort(BaseIOPort):
-    def __init__(self, ports):
+    def __init__(self, ports, yield_ports=False):
         BaseIOPort.__init__(self, 'multi')
-        self.ports = ports
+        self.ports = list(ports)
+        self.yield_ports = yield_ports
 
     def _send(self, message):
         for port in self.ports:
             if not port.closed:
-                # Todo: what if a SocketPort connection closes in-between here?
+                # TODO: what if a SocketPort connection closes in-between here?
                 port.send(message)
 
     def _receive(self, block=True):
-        ports = list(self.ports)
-        random.shuffle(ports)
-        for port in ports:
-            if not port.closed:
-                for message in port.iter_pending():
-                    self._messages.append(message)
+        self._messages.extend(multi_receive(self.ports,
+                                            yield_ports=self.yield_ports,
+                                            block=block))
 
 
-def multi_receive(ports, yield_ports=False):
+def multi_receive(ports, yield_ports=False, block=True):
     """Receive messages from multiple ports.
 
     Generates messages from ever input port. The ports are polled in
     random order for fairness, and all messages from each port are
     yielded before moving on to the next port.
-    
+
     If yield_ports=True, (port, message) is yielded instead of just
     the message.
+
+    If block=False only pending messages will be yielded.
     """
     ports = list(ports)
-    while 1:
+    while True:
+        # Make a shuffled copy of the port list.
         random.shuffle(ports)
-        
-        for port in ports:
-            for message in port.iter_pending():
-                if yield_ports:
-                    yield (port, message)
-                else:
-                    yield message
 
-        sleep()
+        for port in ports:
+            if not port.closed:
+                for message in port.iter_pending():
+                    if yield_ports:
+                        yield port, message
+                    else:
+                        yield message
+
+        if block:
+            sleep()
+        else:
+            break
 
 
 def multi_iter_pending(ports, yield_ports=False):
     """Iterate through all pending messages in ports.
 
-    ports is an iterable of message ports to check.
-
-    This can be used to receive messages from a set of ports in a
-    non-blocking manner.
-
-    If yield_ports=True, (port, message) is yielded instead of just
-    the message.
+    This is the same as calling multi_receive(ports, block=False).
+    The function is kept around for backwards compatability.
     """
-    ports = list(ports)
-    random.shuffle(ports)
+    return multi_receive(ports, yield_ports=yield_ports, block=False)
+
+
+def multi_send(ports, msg):
+    """Send message on all ports."""
     for port in ports:
-        for message in port.iter_pending():
-            if yield_ports:
-                yield (port, message)
-            else:
-                yield message
+        port.send(msg)
